"""Common utilities"""
def safe_get(dct, *keys):
	"""Extract value from nested dictionary
	Args:
		dct (dict): dictionary one want to extrat value from
		*keys (string|list|tuple): keys to exatract value

	Returns:
		value
	"""
	for key in keys:
		try:
			dct = dct[key]
		except:
			return None
	return dct



"""Metta-data (https://github.com/dssg/metta-data) utilities"""

from datetime import date
from calendar import monthrange
import metta
import pandas as pd
from random import randint

def quarter_boundaries(quarter):
    """Returns first and last day of a quarter

    Args:
        quarter (str) quarter, in format '2015Q1'

    Returns: (tuple) datetime.dates for the first and last days of the quarter
    """
    year, quarter = quarter.split('Q')
    year = int(year)
    quarter = int(quarter)
    first_month_of_quarter = 3 * quarter - 2
    last_month_of_quarter = 3 * quarter
    first_day = date(year, first_month_of_quarter, 1)
    last_day = date(year, last_month_of_quarter, monthrange(year, last_month_of_quarter)[1])
    return first_day, last_day


def metta_config(quarter, num_dimensions):
    """Returns metta metadata for a quarter's SOC code classifier matrix

    Args:
        quarter (str) quarter, in format '2015Q1'
        num_dimensions (int) Number of features in matrix

    Returns: (dict) metadata suitable for metta.archive_train_test
    """
    first_day, last_day = quarter_boundaries(quarter)
    return {
        'start_time': first_day,
        'end_time': last_day,
        'prediction_window': 3, # ???
        'label_name': 'onet_soc_code',
        'label_type': 'categorical',
        'matrix_id': 'job_postings_{}'.format(quarter),
        'feature_names': ['doc2vec_{}'.format(i) for i in range(num_dimensions)],
    }

def upload_to_metta(train_features_path, train_labels_path, test_features_path, test_labels_path, train_quarter, test_quarter, num_dimensions):
    """Store train and test matrices using metta

    Args:
        train_features_path (str) Path to matrix with train features
        train_labels_path (str) Path to matrix with train labels
        test_features_path (str) Path to matrix with test features
        test_labels_path (str) Path to matrix with test labels
        train_quarter (str) Quarter of train matrix
        test_quarter (str) Quarter of test matrix
        num_dimensions (int) Number of features
    """
    train_config = metta_config(train_quarter, num_dimensions)
    test_config = metta_config(test_quarter, num_dimensions)

    X_train = pd.read_csv(train_features_path, sep=',')
    X_train.columns = ['doc2vec_'+str(i) for i in range(X_train.shape[1])]
    #X_train['label'] = pd.Series([randint(0,23) for i in range(len(X_train))])
    Y_train = pd.read_csv(train_labels_path)
    Y_train.columns = ['onet_soc_code']
    train = pd.concat([X_train, Y_train], axis=1)

    X_test = pd.read_csv(test_features_path, sep=',')
    X_test.columns = ['doc2vec_'+str(i) for i in range(X_test.shape[1])]
    #X_test['label'] = pd.Series([randint(0,23) for i in range(len(X_test))])
    Y_test = pd.read_csv(test_labels_path)
    Y_test.columns = ['onet_soc_code']
    test = pd.concat([X_test, Y_test], axis=1)
    #print(train.head())
    #print(train.shape)
    #print(test.head())
    #print(test.shape)
    metta.archive_train_test(
        train_config,
        X_train,
        test_config,
        X_test,
        directory='wdi'
    )

if __name__ == '__main__':
    upload_to_metta('../tmp/job_features_train_2011Q1.csv',
                    '../tmp/job_labels_train_2011Q1.csv',
                    '../tmp/job_features_test_2016Q1.csv',
                    '../tmp/job_labels_test_2016Q1.csv',
                    '2011Q1',
                    '2016Q1',
                    500)



"""Elasticsearch utilities"""

from elasticsearch import Elasticsearch, TransportError
from elasticsearch.client import IndicesClient
from elasticsearch.helpers import streaming_bulk
import contextlib
import logging
import os
import time
import uuid


HOSTNAME = os.getenv('ELASTICSEARCH_ENDPOINT', 'localhost:9200')


def basic_client():
    """Returns an Elasticsearch basic client that is responsive
    to the environment variable ELASTICSEARCH_ENDPOINT"""
    es_connected = False
    while not es_connected:
        try:
            ES = Elasticsearch(
                hosts=[HOSTNAME]
            )
            es_connected = True
        except TransportError as e:
            logging.info('Not yet connected: %s, sleeping for 1s', e)
            time.sleep(1)
    return ES


def indices_client():
    """Returns an Elasticsearch indices client that is responsive
    to the environment variable ELASTICSEARCH_ENDPOINT"""
    es_connected = False
    while not es_connected:
        try:
            ES = Elasticsearch(
                hosts=[HOSTNAME]
            )
            es_connected = True
        except TransportError as e:
            logging.info('Not yet connected: %s, sleeping for 1s', e)
            time.sleep(1)
    return IndicesClient(ES)


def create_index(index_name, index_config, client):
    """Creates an index with a given configuration

    Args:
        index_name (str): Name of the index you want to create
        index_config (dict) configuration for the index
        client (Elasticsearch.IndicesClient) the Elasticsearch client
    """
    client.create(index=index_name, body=index_config)


def get_index_from_alias(alias_name, index_client=None):
    """Retrieve the base index name from an alias

    Args:
        alias_name (str) Name of the alias
        index_client (Elasticsearch.IndicesClient) an Elasticsearch index
            client. Optional, will create one if not given

    Returns: (str) Name of index
    """
    index_client = index_client or indices_client()
    if not index_client.exists_alias(name=alias_name):
        return None
    return list(index_client.get_alias(name=alias_name).keys())[0]


def atomic_swap(alias_name, new_index_name, index_client):
    """Points an alias to a new index, then delete the old index if needed

    Uses client.update_aliases to perform this with zero downtime

    Args:
        alias_name (str) Name of the alias
        new_index_name (str) The new index that the alias should point to
        index_client (Elasticsearch.IndicesClient) Elasticsearch index client
    """
    logging.info('Performing atomic index alias swap')
    if index_client.exists_alias(name=alias_name):
        old_index_name = get_index_from_alias(alias_name, index_client)
        logging.info('Removing old as well as adding new')
        actions = {'actions': [
            {'remove': {'index': old_index_name, 'alias': alias_name}},
            {'add': {'index': new_index_name, 'alias': alias_name}}
        ]}
        index_client.update_aliases(body=actions)
        index_client.delete(index=old_index_name)
    else:
        logging.info('Old alias not found, only adding new')
        actions = {'actions': [
            {'add': {'index': new_index_name, 'alias': alias_name}}
        ]}
        index_client.update_aliases(body=actions)


@contextlib.contextmanager
def zero_downtime_index(index_name, index_config):
    """Context manager to create a new index based on a given alias,
    allow the caller to index it, and then point the alias to the new index

    Args:
        index_name (str) Name of an alias that should point to the new index
        index_config (dict) Configuration for the new index

    Yields: (name) The full name of the new index
    """
    client = indices_client()
    temporary_name = index_name + '_' + str(uuid.uuid4())
    logging.info('creating index with config %s', index_config)
    create_index(temporary_name, index_config, client)
    try:
        yield temporary_name
        atomic_swap(index_name, temporary_name, client)
    except Exception:
        logging.error(
            'deleting temporary index %s due to error:',
            temporary_name,
            exc_info=True
        )
        client.delete(index=temporary_name)


class ElasticsearchIndexerBase(object):
    def __init__(self, s3_conn, es_client):
        """
        Base class for Elasticsearch indexers

        Subclasses implement the index setting definition and transformation of data,
        The base class handles index management and bulk indexing with ES

        Args:
            s3_conn - a boto s3 connection
            es_client - an Elasticsearch indices client
        """
        self.s3_conn = s3_conn
        self.es_client = es_client

    def index_config(self):
        """Combines setting and mapping config into a full index configuration
        Returns: dict
        """
        return {
            'settings': self.settings,
            'mappings': self.mappings
        }

    def replace(self):
        """Replace index with a new one
        zero_downtime_index for safety and rollback
        """
        with zero_downtime_index(self.alias_name, self.index_config()) as target_index:
            self.index_all(target_index)

    def append(self):
        """Index documents onto an existing index"""
        target_index = get_index_from_alias(self.alias_name)
        if not target_index:
            self.replace()
        else:
            self.index_all(target_index)

    def index_all(self, index_name):
        """Index all available documents, using streaming_bulk for speed
        Args:

        index_name (string): The index
        """
        oks = 0
        notoks = 0
        for ok, item in streaming_bulk(
            self.es_client,
            self._iter_documents(index_name)
        ):
            if ok:
                oks += 1
            else:
                notoks += 1
        logging.info(
            "Import results: %d ok, %d not ok",
            oks,
            notoks
        )



"""Iteration utilities"""

class Batch:
    """Yields batches (groups) from an iterable

    Modified from:
    http://codereview.stackexchange.com/questions/118883/split-up-an-iterable-into-batches

    Args:
        iterable (iterable) any iterable
        limit (int) How many items to include per group
    """
    def __init__(self, iterable, limit=None):
        self.iterator = iter(iterable)
        self.limit = limit
        try:
            self.current = next(self.iterator)
        except StopIteration:
            self.on_going = False
        else:
            self.on_going = True

    def group(self):
        """Yield a group from the iterable"""
        yield self.current
        # start enumerate at 1 because we already yielded the last saved item
        for num, item in enumerate(self.iterator, 1):
            self.current = item
            if num == self.limit:
                break
            yield item
        else:
            self.on_going = False

    def __iter__(self):
        """Implementation of __iter__ to allow a standard interface:

        for group in Batch(iterable, 10):
            do_stuff(group)
        """

        while self.on_going:
            yield self.group()



"""Testing utilities"""
from skills_utils.job_posting_import import JobPostingImportBase

import unittest

MANDATORY_FIELDS = [
    'title',
    'description',
    'datePosted',
    'validThrough',
    'jobLocation',
]


class SampleImporter(JobPostingImportBase):
    """A basic importer that can be run through the ImporterTest"""
    def _iter_postings(self, quarter):
        pass

    def _transform(self, document):
        transformed = {field: '' for field in MANDATORY_FIELDS}
        transformed['@context'] = "http://schema.org"
        transformed['@type'] = 'JobPosting'
        return transformed


class ImporterTest(unittest.TestCase):
    """Common superclass for all partner ETL tests"""
    importer_class = SampleImporter
    sample_input_document = {}

    def test_schema_org(self):
        """Make basic assertions about common schema"""

        assert hasattr(self, 'sample_input_document')
        importer = self.importer_class(partner_id='xx')
        transformed = importer._transform(self.sample_input_document)
        assert transformed['@context'] == 'http://schema.org'
        assert transformed['@type'] == 'JobPosting'
        for field in MANDATORY_FIELDS:
            assert field in transformed



"""I/O utilities"""
import json
import logging


def stream_json_file(local_file):
    """Stream a JSON file (in JSON-per-line format)

    Args:
        local_file (file-like object) an open file-handle that contains a
            JSON string on each line
    Yields:
        (dict) JSON objects
    """
    for i, line in enumerate(local_file):
        try:
            data = json.loads(line.decode('utf-8'))
            yield data
        except ValueError as e:
            logging.warning("Skipping line %d due to error: %s", i, e)
            continue



"""Common Schema Job Posting utilities"""
import logging


class JobPostingImportBase(object):
    """Base class for extracting and transforming job postings from
    some source into a common schema. (http://schema.org/JobPosting)

    Subclasses must implement _id, _iter_postings, and _transform.

    Args:
        partner_id (str) An short identifier for the partner (ie NLX, VA)
        s3_conn (boto.s3.connection) an s3 connection, if needed
        onet_cache (skills_ml.onet_cache.OnetCache) a wrapper for ONET data, if needed
    """
    def __init__(self, partner_id, s3_conn=None, onet_cache=None):
        self.partner_id = partner_id
        self.s3_conn = s3_conn
        self.onet_cache = onet_cache

    def postings(self, quarter, stats_counter=None):
        """Yield job postings in common schema format

        Args:
            quarter (str) The quarter, in format '2015Q1'
            stats_counter (object, optional) A counter that can track both
                input and output documents using a 'track' method.
        """
        logging.info('Finding postings for %s', quarter)
        for posting in self._iter_postings(quarter):
            transformed = self._transform(posting)
            transformed['id'] = '{}_{}'.format(
                self.partner_id,
                self._id(posting)
            )
            if stats_counter:
                stats_counter.track(
                    input_document=posting,
                    output_document=transformed
                )
            yield transformed

    def _id(self, document):
        """Given a document, compute a source-specific id for the job posting.
        To be implemented by subclasses

        Args:
            document - The document, in original form

        Returns: (str) an id for the document
        """
        pass

    def _iter_postings(self, quarter):
        """Given a quarter, yield all relevant raw job posting documents.
        To be implemented by subclasses

        Args:
            quarter (str) The quarter, in format '2015Q1'

        Yields: job posting documents in their original format
        """
        pass

    def _transform(self, document):
        """Given a job posting document, transform it into the common schema.
        (http://schema.org/JobPosting)
        To be implemented by subclasses.

        Args:
            document - The document, in original form

        Returns: (dict) The job posting, in common schema form
        """
        pass



"""Filesystem-related utilities"""

from functools import wraps
import os
import json

CACHE_DIRECTORY = 'tmp/'


def cache_json(filename):
    """Caches the JSON-serializable output of the function to a given file

    Args:
        filename (str) The filename (sans directory) to store the output

    Returns: decorator, applicable to a function that produces JSON-serializable output
    """
    def cache_decorator(cacheable_function):
        @wraps(cacheable_function)
        def cache_wrapper(*args, **kwargs):
            path = CACHE_DIRECTORY + filename
            check_create_folder(path)
            if os.path.exists(path):
                with open(path) as infile:
                    return json.load(infile)
            else:
                function_output = cacheable_function(*args, **kwargs)
                with open(path, 'w') as outfile:
                    json.dump(function_output, outfile)
                return function_output
        return cache_wrapper
    return cache_decorator


def check_create_folder(filename):
    """Check if the folder exisits. If not, create the folder"""
    os.makedirs(os.path.dirname(filename), exist_ok=True)



import hashlib


def md5(string):
    """Returns the md5 hash of a string

    Args:
        string (str) any string

    Returns: (str) the md5 hash
    """
    return hashlib.md5(string.encode('utf-8')).hexdigest()



from skills_utils.io import stream_json_file
from skills_utils.iteration import Batch
from skills_utils.job_posting_import import JobPostingImportBase
from skills_utils.s3 import split_s3_path
from skills_utils.time import datetime_to_quarter, overlaps, quarter_to_daterange
from skills_utils.common import safe_get



#!/usr/bin/env python

from setuptools import find_packages, setup

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.readlines()

with open('requirements_dev.txt') as requirements_file:
    test_requirements = requirements_file.readlines() + requirements

setup(
    name='Skills Utils',
    version='0.4.0',
    description='Shared utilities for skills/jobs taxonomy creation',
    author="Center for Data Science and Public Policy",
    author_email='datascifellows@gmail.com',
    url='https://github.com/workforce-data-initiative/skills-utils',
    packages=find_packages(include=['skills_utils*']),
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



"""
Common S3 utilities
"""
import boto
import json
import logging
import os
from collections.abc import MutableMapping

import s3fs


def split_s3_path(path):
    """
    Args:
        path: (str) an s3 path including a bucket
            (bucket_name/prefix/prefix2)
    Returns:
        A tuple containing the bucket name and full prefix)
    """
    return path.split('/', 1)


def upload(s3_conn, filepath, s3_path):
    """Uploads the given file to s3

    Args:
        s3_conn: (boto.s3.connection) an s3 connection
        filepath (str) the local filename
        s3_path (str) the destination path on s3
    """
    bucket_name, prefix = split_s3_path(s3_path)
    bucket = s3_conn.get_bucket(bucket_name)
    filename = os.path.basename(filepath)

    key = boto.s3.key.Key(
        bucket=bucket,
        name='{}/{}'.format(prefix, filename)
    )
    logging.info('uploading from %s to %s', filepath, key)
    key.set_contents_from_filename(filepath)


def upload_dict(s3_conn, s3_prefix, data_to_sync):
    """Syncs a dictionary to an S3 bucket, serializing each value in the
    dictionary as a JSON file with the key as its name.

    Args:
        s3_conn: (boto.s3.connection) an s3 connection
        s3_prefix: (str) the destination prefix
        data_to_sync: (dict)
    """
    bucket_name, prefix = split_s3_path(s3_prefix)
    bucket = s3_conn.get_bucket(bucket_name)

    for key, value in data_to_sync.items():
        full_name = '{}/{}.json'.format(prefix, key)
        s3_key = boto.s3.key.Key(
            bucket=bucket,
            name=full_name
        )
        logging.info('uploading key %s', full_name)
        s3_key.set_contents_from_string(json.dumps(value))


def download(s3_conn, out_filename, s3_path):
    """Downloads the given s3_path

    Args:
        s3_conn (boto.s3.connection) a boto s3 connection
        out_filename (str) local filename to save the file
        s3_path (str) the source path on s3
    """
    bucket_name, prefix = split_s3_path(s3_path)
    bucket = s3_conn.get_bucket(bucket_name)
    key = boto.s3.key.Key(
        bucket=bucket,
        name=prefix
    )
    logging.info('loading from %s into %s', key, out_filename)
    key.get_contents_to_filename(out_filename, cb=log_download_progress)


def log_download_progress(num_bytes, obj_size):
    """Callback that boto can use to log download or upload progress"""
    logging.info('%s bytes transferred out of %s total', num_bytes, obj_size)


def list_files(s3_conn, s3_path):
    bucket_name, prefix = split_s3_path(s3_path)
    bucket = s3_conn.get_bucket(bucket_name)
    key = boto.s3.key.Key(
        bucket=bucket,
        name=prefix
    )
    files = []
    for key in bucket.list(prefix=prefix):
        files.append(key.name.split('/')[-1])
    return list(filter(None, files))


class S3BackedJsonDict(MutableMapping):
    """A JSON-serializable dictionary that is backed by S3.

    Not guaranteed to be thread or multiprocess-safe - An attempt is made before saving to merge
    local changes with others that may have happened to the S3 file since this object was loaded,
    but this is not atomic.
    It is recommended that only one version of a given file be modified at a time.

    Will periodically save, but users must call .save() before closing to save all changes.

    Keyword Args:
        path (string): A full s3 path, including bucket (but without the .json suffix),
            used for saving the dictionary.
    """
    SAVE_EVERY_N_UPDATES = 1000

    def __init__(self, *args, **kw):
        self.path = kw.pop('path') + '.json'
        self.fs = s3fs.S3FileSystem()

        if self.fs.exists(self.path):
            with self.fs.open(self.path, 'rb') as f:
                data = f.read().decode('utf-8') or '{}'
                self._storage = json.loads(data)
        else:
            self._storage = dict()

        self.num_updates = 0
        logging.info('Loaded storage with %s keys', len(self))

    def __getitem__(self, key):
        return self._storage[key]

    def __iter__(self):
        return iter(self._storage)

    def __len__(self):
        return len(self._storage)

    def __delitem__(self, key):
        del self._storage[key]

    def __setitem__(self, key, value):
        self._storage[key] = value
        self.num_updates += 1
        if self.num_updates % self.SAVE_EVERY_N_UPDATES == 0:
            logging.info('Auto-saving after %s updates', self.num_updates)
            self.save()

    def __keytransform__(self, key):
        return key

    def __contains__(self, key):
        return key in self._storage

    def save(self):
        logging.info('Attempting to save storage of length %s to %s', len(self), self.path)
        if self.fs.exists(self.path):
            with self.fs.open(self.path, 'rb') as f:
                saved_string = f.read().decode('utf-8') or '{}'
                saved_data = json.loads(saved_string)
                logging.info(
                    'Merging %s in-memory keys with %s stored keys. In-memory data takes priority',
                    len(self),
                    len(saved_data)
                )
                saved_data.update(self._storage)
                self._storage = saved_data
        with self.fs.open(self.path, 'wb') as f:
            f.write(json.dumps(self._storage).encode('utf-8'))



"""Time utilities"""
from datetime import date, timedelta
import math


def quarter_to_daterange(quarter):
    """Convert a quarter in arbitrary filename-ready format (e.g. 2015Q1)
    into start and end datetimes"""
    assert len(quarter) == 6
    year = int(quarter[0:4])
    quarter = quarter[5]
    MONTH_DAY = {
        '1': ((1, 1), (3, 31)),
        '2': ((4, 1), (6, 30)),
        '3': ((7, 1), (9, 30)),
        '4': ((10, 1), (12, 31))
    }
    md = MONTH_DAY[quarter]
    start_md, end_md = md
    return (
        date(year, *start_md),
        date(year, *end_md)
    )


def datetime_to_year_quarter(dt):
    """
    Args:
        dt: a datetime
    Returns:
        tuple of the datetime's year and quarter
    """
    year = dt.year
    quarter = int(math.ceil(float(dt.month)/3))
    return (year, quarter)


def datetime_to_quarter(dt):
    """
    Args:
        dt: a datetime
    Returns:
        the datetime's quarter in string format (2015Q1)
    """
    year, quarter = datetime_to_year_quarter(dt)
    return '{}Q{}'.format(year, quarter)


def overlaps(start_one, end_one, start_two, end_two):
    return start_one <= end_two and start_two <= end_one


def dates_in_range(start_date, end_date):
    """Returns all dates between two dates.

    Inclusive of the start date but not the end date.

    Args:
        start_date (datetime.date)
        end_date (datetime.date)

    Returns:
        (list) of datetime.date objects
    """
    return [
        start_date + timedelta(n)
        for n in range(int((end_date - start_date).days))
    ]
